// Include the GUI:
#define IMGUI_DEFINE_MATH_OPERATORS
#include <imgui_internal.h>
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

#include <glad/glad.h>
#include <GLFW/glfw3.h>
#include <glm/gtx/transform.hpp>
#include <glm/gtx/orthonormalize.hpp>

#include <iostream>
#include <memory>
#include <cmath>

#include "Shader.h"
#include "Mesh.h"
#include "Body.h"
#include "Scene.h"
#include "PlaneCube.h"

#define SUBDIVISIONS 1
static const int OtherBodyDim = 0;

#if SUBDIVISIONS <= 1
static const std::string ObjFile = std::string( std::string(CMAKE_SOURCE_DIR)+std::string("/models/ico1.obj") );
#elif SUBDIVISIONS <= 2
static const std::string ObjFile = std::string( std::string(CMAKE_SOURCE_DIR)+std::string("/models/ico2.obj") );
#elif SUBDIVISIONS <= 3
static const std::string ObjFile = std::string( std::string(CMAKE_SOURCE_DIR)+std::string("/models/ico3.obj") );
#else
static const std::string ObjFile = std::string( std::string(CMAKE_SOURCE_DIR)+std::string("/models/ico4.obj") );
#endif

/**
 * @brief Convert a given 2D screen position to a 3D position in view space.
 */
glm::vec3 screenToTrackball( ImVec2 screenPos, ImVec2 screenSize ) {
    auto x = (screenPos.x-(screenSize.x/2.f)) / (screenSize.x/2.f);
    auto y = ((screenSize.y/2.f)-screenPos.y) / (screenSize.y/2.f);
    auto z = 1 - x*x - y*y;
    glm::vec3 trackballPos;
    if (z < 0)
        trackballPos = glm::vec3( x/sqrtf(x*x+y*y), y/sqrtf(x*x+y*y), 0 );
    else
        trackballPos = glm::vec3( x, y, sqrtf(z) );
    return trackballPos;
}

/**
 * Initializes the application including the GUI
 */
int main(int, char**)
{
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(1300, 1300, "Advanced Computergraphics: Collision Detection", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);

    // Initialize GLAD functions:
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;

    // Setup Dear ImGui style:
    ImGui::StyleColorsClassic();

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Defines the with of the left menu:
    int menuWidth = 260;

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf", 24.0f);

    // Enable depth test:
    glEnable(GL_DEPTH_TEST);

    // Create our scene
    auto scene = std::make_shared<Scene>();
    auto body0 = std::make_shared<Body>( ObjFile.c_str() );
    scene->objects.push_back( body0 );
    scene->bodies.push_back( body0 );
    scene->centerBody = body0;
    static int OtherBodyCount = (int)pow(OtherBodyDim,3);
    static int OtherBodyCountX = std::max(static_cast<int>(std::floor(cbrt(OtherBodyCount))),1);
    static int OtherBodyCountY = std::max(static_cast<int>(std::floor(cbrt(OtherBodyCount))),1);
    static int OtherBodyCountZ = std::max(static_cast<int>(std::floor(cbrt(OtherBodyCount))),1);
    if (OtherBodyDim == 0)
        OtherBodyCountX = 1;
    else if (OtherBodyDim == 1)
        OtherBodyCountX = 2;
    auto pos = glm::vec3(0,0,0);
    auto color = glm::vec4(0,0,0,1);
    static const float RangeStart = -2;
    static const float RangeEnd = 2;
    for (int x=0; x<OtherBodyCountX; ++x) {
        float fx = ((float)x/(float)(OtherBodyCountX));
        color.r = fx + (1.f/((float)OtherBodyCountX));
        if (OtherBodyCountX==1)
            pos.x = RangeStart/2.0;
        else
            pos.x = RangeStart + (RangeEnd-(RangeStart))*fx + (std::abs(RangeEnd)/((float)OtherBodyCountX));
        for (int y=0; y<OtherBodyCountY; ++y) {
            float fy = ((float)y/(float)(OtherBodyCountY));
            color.g = fy + (1/((float)OtherBodyCountY));
            if (OtherBodyCountY==1)
                pos.y = RangeStart/2.f;
            else
                pos.y = RangeStart + (RangeEnd-(RangeStart))*fy + (std::abs(RangeEnd)/((float)OtherBodyCountY));
            for (int z=0; z<OtherBodyCountZ; ++z) {
                float fz = ((float)z/(float)(OtherBodyCountZ));
                color.b = fz + (1/((float)OtherBodyCountZ));
                if (OtherBodyCountZ==1)
                    pos.z = RangeStart/2.f;
                else
                    pos.z = RangeStart + (RangeEnd-(RangeStart))*fz + (std::abs(RangeEnd)/((float)OtherBodyCountZ));
                if (std::abs(pos.x)<=0.001f && std::abs(pos.y)<=0.001f && std::abs(pos.z)<=0.001f)
                    continue;
                auto body = std::make_shared<Body>( ObjFile.c_str() );
                scene->objects.push_back( body );
                scene->bodies.push_back( body );
                scene->orbitingBodies.push_back( body );
                body->setTranslation( pos );
                body->m_color = color;
                body->setScale( glm::vec3(0.2f) );
            }
        }
    }

    scene->centerBody->m_color = glm::vec4(0.3,0.3,0.3,1);
    scene->centerBody->setTranslation( glm::vec3(0,0,0) );
    scene->centerBody->setScale( glm::vec3(0.3f) );

    auto planeCube = std::make_shared<PlaneCube>(scene);
    scene->objects.push_back( planeCube );

    Shader shader(CMAKE_SOURCE_DIR "/shader/phong.vs", CMAKE_SOURCE_DIR "/shader/phong.fs");

    glm::vec3 lastTrackballPos1(0,0,1);
    glm::vec3 lastTrackballPos2(0,0,1);

    // Should use perspective projection?:
    float ZoomFactor = 70;
    glm::vec3 eye(0,0,10);
    glm::mat3 M_trackball_objects = glm::identity<glm::mat3>();
    glm::mat3 M_trackball_cube = glm::identity<glm::mat3>();
    bool leftMouseUpBefore = true;
    bool rightMouseUpBefore = true;

    glm::mat4 view = glm::lookAt(eye, glm::vec3(0,0,0), glm::vec3(0,1,0));

    // Main loop which is executed every frame until the window is closed:
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

        // Create the GUI:
        {
            // Start the Dear ImGui frame
            ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplGlfw_NewFrame();
            ImGui::NewFrame();
        }

        // Setup the GL viewport
        glViewport(0, 0, display_w, display_h);
        glClearColor(0, 0, 0, 1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Calculate aspect ratio (might be used for perspective transformation):
        float aspectRatio = float(display_w) / display_h;

        // *** Virtual Trackball ***

        M_trackball_cube = glm::identity<glm::mat4>();
        M_trackball_objects = glm::identity<glm::mat4>();
        bool leftMouseDown = ImGui::IsMouseDown(ImGuiMouseButton_Left);
        bool rightMouseDown = ImGui::IsMouseDown(ImGuiMouseButton_Right);
        if (!ImGui::GetIO().WantCaptureMouse) {
            auto m = ImGui::GetMousePos();
            auto VfW = view;
            auto WfV = glm::inverse(view);
            auto trackballPos_screen = screenToTrackball(m, ImVec2((float)display_w,(float)display_h));

            // Objects are rotated in world space
            auto trackballPos1 = glm::vec3( WfV * glm::vec4(trackballPos_screen,1) );

            // Cube is rotated in screen space
            auto trackballPos2 = trackballPos_screen;

            // Left mouse button changed from up to down?
            if (leftMouseDown) {
                if (leftMouseUpBefore)
                    lastTrackballPos1 = trackballPos1;
                leftMouseUpBefore = false;
            } else
                leftMouseUpBefore = true;

            // Right mouse button changed from up to down?
            if (rightMouseDown) {
                if (rightMouseUpBefore)
                    lastTrackballPos2 = trackballPos2;
                rightMouseUpBefore = false;
            } else
                rightMouseUpBefore = true;

            // Calculate virtual trackball rotation
            if (leftMouseDown) {
                auto r1 = glm::vec4(glm::cross(lastTrackballPos1, trackballPos1),0);
                float a1 = acosf(glm::dot(glm::normalize(lastTrackballPos1),glm::normalize(trackballPos1))) * 80.f;
                if (glm::length(r1)>0.0001f && fabsf(a1)>0.0001f) {
                    M_trackball_objects = glm::rotate(a1,glm::vec3(r1));
                    lastTrackballPos1 = trackballPos1;
                }
            }
            if (rightMouseDown) {
                auto r2 = glm::vec4(glm::cross(lastTrackballPos2, trackballPos2),0);
                float a2 = acosf(glm::dot(glm::normalize(lastTrackballPos2),glm::normalize(trackballPos2))) * 4.f;
                if (glm::length(r2)>0.0001f && fabsf(a2)>0.0001f) {
                    M_trackball_cube = glm::rotate(a2,glm::vec3(r2));
                    lastTrackballPos2 = trackballPos2;
                }
            }
        }

        // Rotate view matrix
        view = glm::translate(-eye) * glm::mat4(M_trackball_cube) * glm::translate(eye) * view;

        // Update camera distance when scrolling the mouse:
        if(ImGui::GetIO().MouseWheel != 0)
            ZoomFactor = std::min<float>(std::max<float>(ZoomFactor + (ImGui::GetIO().MouseWheel > 0 ? -5 : 5), 1), 179);

        // Rotate oribiting bodies
        for (auto body : scene->orbitingBodies) {
            auto pos = body->getPosition();
            auto M_WfB_R = body->getRotation();
            body->setRotation( glm::translate(-pos) * glm::mat4(M_trackball_objects) * glm::translate(pos) * M_WfB_R );
        }

        glm::mat4 projection = glm::perspective( glm::radians(ZoomFactor), aspectRatio, 0.01f, 1000.f);
        shader.setUniform("projection", projection);
        shader.setUniform("view", view);
        shader.setUniform("viewPos", M_trackball_cube*eye);
        shader.setUniform("lightColor", glm::vec3(1,1,1));
        shader.setUniform("lightPos", glm::vec3(0,0,10.f-4.f));

        // Timing window
        static ImGuiWindowFlags flags = ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoSavedSettings | ImGuiWindowFlags_NoResize;
        ImGui::SetNextWindowPos(ImVec2(0,0));
        ImGui::SetNextWindowSize(ImVec2(300,80));
        if (ImGui::Begin("Timing", nullptr, flags)) {
            ImGui::Text("%u Hz [%f us]", planeCube->m_fps, planeCube->m_sepPlaneDurationMicroSec );
            ImGui::End();
        }

        // Render the meshes
        for (auto obj : scene->objects)
            obj->render( shader );

        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}